@@ -5,18 +5,23 @@ package DBIx::DataModel::Statement;
 
 use warnings;
 use strict;
-use Carp;
-use List::Util       qw/min first/;
+use List::Util       qw/min/;
 use List::MoreUtils  qw/firstval any/;
-use Scalar::Util     qw/weaken refaddr reftype dualvar/;
-use Storable         qw/dclone freeze/;
-use Params::Validate qw/validate ARRAYREF HASHREF/;
-use POSIX            qw/INT_MAX/;
-use Acme::Damn       qw/damn/;
-use namespace::autoclean;
+use Scalar::Util     qw/weaken dualvar/;
+use POSIX            qw/LONG_MAX/;
+use Clone            qw/clone/;
+use Carp::Clan       qw[^(DBIx::DataModel::|SQL::Abstract)];
+use Try::Tiny        qw/try catch/;
+use Module::Load     qw/load/;
+use MRO::Compat;
 
 use DBIx::DataModel;
-{no strict 'refs'; *CARP_NOT = \@DBIx::DataModel::CARP_NOT;}
+use DBIx::DataModel::Meta::Utils qw/define_readonly_accessors does/;
+use namespace::clean;
+
+#----------------------------------------------------------------------
+# internals
+#----------------------------------------------------------------------
 
 use overload
 
@@ -24,7 +29,7 @@ use overload
   # also useful to show the SQL (if in sqlized state)
   '""' => sub {
     my $self = shift;
-    my $string = eval {my ($sql, @bind) = $self->sql;
+    my $string = try {my ($sql, @bind) = $self->sql;
                        __PACKAGE__ . "($sql // " . join(", ", @bind) . ")"; }
               || overload::StrVal($self);
   }
@@ -47,22 +52,22 @@ use constant {
 #----------------------------------------------------------------------
 
 sub new {
-  my ($class, $connected_source, %other_args) = @_;
+  my ($class, $source, %other_args) = @_;
 
-  # check $connected_source
-  $connected_source 
-    && $connected_source->isa('DBIx::DataModel::ConnectedSource')
-    or croak "invalid connected_source for DBIx::DataModel::Statement->new()";
+  # check $source
+  $source 
+    && $source->isa('DBIx::DataModel::Source')
+    or croak "invalid source for DBIx::DataModel::Statement->new()";
 
   # build the object
   my $self = bless {status           => NEW,
                     args             => {},
                     pre_bound_params => {},
                     bound_params     => [],
-                    connected_source => $connected_source}, $class;
+                    source           => $source}, $class;
 
   # add placeholder_regex
-  my $prefix = $connected_source->schema->{placeholder_prefix};
+  my $prefix = $source->schema->{placeholder_prefix};
   $self->{placeholder_regex} = qr/^\Q$prefix\E(.+)/;
 
   # parse remaining args, if any
@@ -73,37 +78,31 @@ sub new {
 
 
 # accessors
-DBIx::DataModel::Meta::Utils->define_readonly_accessors(
-  __PACKAGE__, qw/connected_source status/,
-);
-sub meta_source {shift->{connected_source}->meta_source}
-sub schema      {shift->{connected_source}->schema}
+define_readonly_accessors( __PACKAGE__, qw/source status/);
+
+# proxy methods
+sub meta_source {shift->{source}->metadm}
+sub schema      {shift->{source}->schema}
 
 
-
-# don't remember why this "clone()" method was ever created.
-# Keep the code  around for a while ...
-
-# sub clone {
-#   my ($self) = @_;
-#   $self->{status} < PREPARED
-#     or croak "can't clone() when in status $self->{status}";
-
-#   return dclone($self); # THINK: should use Clone::clone instead?
-# }
-
-
-# THINK : not documented yet, is this method useful ?
+# back to the original state
 sub reset {
   my ($self, %other_args) = @_;
 
-  my $new = (ref $self)->new($self->{connected_source}, %other_args);
+  my $new = (ref $self)->new($self->{source}, %other_args);
   %$self = (%$new);
 
   return $self;
 }
 
 
+sub arg {
+  my ($self, $arg_name) = @_;
+
+  my $args = $self->{args} || {};
+  return $args->{$arg_name};
+}
+
 
 
 #----------------------------------------------------------------------
@@ -114,8 +113,8 @@ sub reset {
 sub sql {
   my ($self) = @_;
 
-  $self->{status} >= SQLIZED
-    or croak "can't call sql() when in status $self->{status}";
+  $self->status >= SQLIZED
+    or croak "can't call sql() when in status ". $self->status;
 
   return wantarray ? ($self->{sql}, @{$self->{bound_params}})
                    : $self->{sql};
@@ -127,18 +126,19 @@ sub bind {
 
   # arguments can be a list, a hashref or an arrayref
   if (@args == 1) {
-    for (reftype($args[0]) || "") {
-      /^HASH$/  and do {@args = %{$args[0]}; last;};
-      /^ARRAY$/ and do {my $i = 0; @args = map {($i++, $_)} @{$args[0]}; last};
-      #otherwise
+    if (does $args[0], 'HASH') {
+      @args = %{$args[0]};
+    }
+    elsif (does $args[0], 'ARRAY') {
+      my $i = 0; @args = map {($i++, $_)} @{$args[0]};
+    }
+    else {
       croak "unexpected arg type to bind()";
     }
   }
-  elsif (@args == 3) { # name => value, \%args (see L<DBI/bind_param>)
-    my $indices = $self->{param_indices}{$args[0]};
-    my $bind_param_args = pop @args;
-    defined $indices or croak "no such named placeholder : $args[0]";
-    $self->{bind_param_args}[$_] = $bind_param_args foreach @$indices;
+  elsif (@args == 3) { # name => value, \%datatype (see L<DBI/bind_param>)
+    # transform into ->bind($name => [$value, \%datatype])
+    @args = ($args[0], [$args[1], $args[2]]);
   }
   elsif (@args % 2 == 1) {
     croak "odd number of args to bind()";
@@ -146,7 +146,7 @@ sub bind {
 
   # do bind (different behaviour according to status)
   my %args = @args;
-  if ($self->{status} < SQLIZED) {
+  if ($self->status < SQLIZED) {
     while (my ($k, $v) = each %args) {
       $self->{pre_bound_params}{$k} = $v;
     }
@@ -159,6 +159,10 @@ sub bind {
     }
   }
 
+  # THINK : probably we should check here that $args{__schema}, if present,
+  # is the same as $self->schema (same database connection) ... but how
+  # to check for "sameness" on database handles ?
+
   return $self;
 }
 
@@ -166,8 +170,8 @@ sub bind {
 sub refine {
   my ($self, %more_args) = @_;
 
-  $self->{status} <= REFINED
-    or croak "can't refine() when in status $self->{status}";
+  $self->status <= REFINED
+    or croak "can't refine() when in status " . $self->status;
   $self->{status} = REFINED;
 
   my $args = $self->{args};
@@ -211,7 +215,7 @@ sub refine {
 
       # -columns : store in $self->{args}{-columns}; can restrict previous list
       /^-columns$/ and do {
-        my @cols = ref $v ? @$v : ($v);
+        my @cols = does($v, 'ARRAY') ? @$v : ($v);
         if (my $old_cols = $args->{-columns}) {
           unless (@$old_cols == 1 && $old_cols->[0] eq '*' ) {
             foreach my $col (@cols) {
@@ -226,15 +230,18 @@ sub refine {
       };
 
 
-      # other args are just stored, will be used later
-      /^-( order_by     | group_by | having    | for
-         | result_as    | post_SQL | pre_exec  | post_exec  | post_bless
-         | limit        | offset   | page_size | page_index
-         | column_types | prepare_attrs        | dbi_prepare_method
-         | _left_cols
+      # other args are just stored, they will be used later
+      /^-( order_by       | group_by  | having    | for
+         | union(?:_all)? | intersect | except    | minus
+         | result_as      | post_SQL  | pre_exec  | post_exec  | post_bless
+         | limit          | offset    | page_size | page_index
+         | column_types   | prepare_attrs         | dbi_prepare_method
+         | _left_cols     | where_on
          )$/x
          and do {$args->{$k} = $v; last SWITCH};
 
+      # TODO : this hard-coded list of args should be more abstract
+
       # otherwise
       croak "invalid arg : $k";
 
@@ -250,8 +257,8 @@ sub refine {
 sub sqlize {
   my ($self, @args) = @_;
 
-  $self->{status} < SQLIZED
-    or croak "can't sqlize() when in status $self->{status}";
+  $self->status < SQLIZED
+    or croak "can't sqlize() when in status ". $self->status;
 
   # merge new args into $self->{args}
   $self->refine(@args) if @args;
@@ -261,18 +268,24 @@ sub sqlize {
   my $meta_source  = $self->meta_source;
   my $source_where = $meta_source->{where};
   my $sql_abstract = $self->schema->sql_abstract;
+  my $result_as    = $args->{-result_as} || "";
+
 
   # build arguments for SQL::Abstract::More
   $self->refine(-where => $source_where) if $source_where;
-  my @args_to_copy = qw/-columns -where -order_by -group_by -having
+  my @args_to_copy = qw/-columns -where
+                        -union -union_all -intersect -except -minus
+                        -order_by -group_by -having
                         -limit -offset -page_size -page_index/;
-  my %sqla_args = (-from         => $meta_source->db_from,
+  my %sqla_args = (-from         => clone($self->source->db_from),
                    -want_details => 1);
-  $args->{$_} and $sqla_args{$_} = $args->{$_} for @args_to_copy;
+  defined $args->{$_} and $sqla_args{$_} = $args->{$_} for @args_to_copy;
   $sqla_args{-columns} ||= $meta_source->default_columns;
+  $sqla_args{-limit}   ||= 1
+    if $result_as eq 'firstrow' && $self->schema->autolimit_firstrow;
 
   # "-for" (e.g. "update", "read only")
-  if (($args->{-result_as}||"") ne 'subquery') {
+  if ($result_as ne 'subquery') {
     if ($args->{-for}) {
       $sqla_args{-for} = $args->{-for};
     }
@@ -281,6 +294,28 @@ sub sqlize {
     }
   }
 
+  # "where_on" : conditions to be added in joins
+  if (my $where_on = $args->{-where_on}) {
+    # retrieve components of the join
+    my ($join_op, $first_table, @other_join_args) = @{$sqla_args{-from}};
+    $join_op eq '-join'
+      or croak "datasource for '-where_on' was not a join";
+    my %by_dest_table = reverse @other_join_args;
+
+    # insert additional conditions into appropriate places
+    while (my ($table, $additional_cond) = each %$where_on) {
+      my $join_cond = $by_dest_table{$table}
+        or croak "-where_on => {'$table' => ..}: this table is not in the join";
+      $join_cond->{condition}
+        = $sql_abstract->merge_conditions($join_cond->{condition},
+                                          $additional_cond);
+    }
+
+    # TODO: should be able to use paths and aliases as keys, instead of 
+    # database table names.
+    # TOCHECK: is this stuff still compatible with the bind() method ?
+  }
+
   # generate SQL
   my $sqla_result = $sql_abstract->select(%sqla_args);
 
@@ -323,10 +358,10 @@ sub prepare {
 
   my $meta_source = $self->meta_source;
 
-  $self->sqlize(@args) if @args or $self->{status} < SQLIZED;
+  $self->sqlize(@args) if @args or $self->status < SQLIZED;
 
-  $self->{status} == SQLIZED
-    or croak "can't prepare() when in status $self->{status}";
+  $self->status == SQLIZED
+    or croak "can't prepare() when in status " . $self->status;
 
   # log the statement and bind values
   $self->schema->_debug("PREPARE $self->{sql} / @{$self->{bound_params}}");
@@ -336,7 +371,9 @@ sub prepare {
   my $method       = $self->{args}{-dbi_prepare_method}
                   || $self->schema->dbi_prepare_method;
   my @prepare_args = ($self->{sql});
-  push @prepare_args, $self->{prepare_attrs} if $self->{prepare_attrs};
+  if (my $prepare_attrs = $self->{args}{-prepare_attrs}) {
+    push @prepare_args, $prepare_attrs;
+  }
   $self->{sth}  = $dbh->$method(@prepare_args);
 
   # new status and return
@@ -345,12 +382,20 @@ sub prepare {
 }
 
 
+sub sth {
+  my ($self) = @_;
+
+  $self->prepare              if $self->status < PREPARED;
+  return $self->{sth};
+}
+
+
 
 sub execute {
   my ($self, @bind_args) = @_;
 
   # if not prepared yet, prepare it
-  $self->prepare              if $self->{status} < PREPARED;
+  $self->prepare              if $self->status < PREPARED;
 
   # TODO: DON'T REMEMBER why the line below was here. Keep it around for a while ...
   push @bind_args, offset => $self->{offset}  if $self->{offset};
@@ -359,7 +404,7 @@ sub execute {
 
   # shortcuts
   my $args = $self->{args};
-  my $sth  = $self->{sth};
+  my $sth  = $self->sth;
 
   # previous row_count, row_num and reuse_row are no longer valid
   delete $self->{reuse_row};
@@ -379,19 +424,9 @@ sub execute {
             . CORE::join(", ", @unbound);
 
   # bind parameters and execute
-  if ($self->{bind_param_args}) { # need to bind one by one because of DBI args
-    my $n_bound_params = @{$self->{bound_params}};
-    for my $i (0 .. $n_bound_params-1) {
-      my @bind = ($i, $self->{bound_params}[$i]);
-      my $bind_args = $self->{bind_param_args}[$i];
-      push @bind, $bind_args   if $bind_args;
-      $sth->bind_param(@bind);
-    }
-    $sth->execute;
-  }
-  else {                          # otherwise just call DBI::execute(...)
-    $sth->execute(@{$self->{bound_params}});
-  }
+  my $sqla = $self->schema->sql_abstract;
+  $sqla->bind_params($sth, @{$self->{bound_params}});
+  $sth->execute;
 
   # post_exec callback
   $args->{-post_exec}->($sth)  if $args->{-post_exec};
@@ -401,103 +436,33 @@ sub execute {
 }
 
 
+
+
+
+my %cache_result_class;
+
 sub select {
   my $self = shift;
 
   $self->refine(@_) if @_;
 
-  my $args = $self->{args}; # all combined args
-
-  my $callbacks = CORE::join ", ", grep {exists $args->{$_}} 
-                                        qw/-pre_exec -post_exec -post_bless/;
+  my $arg_result_as = $self->arg(-result_as) || 'rows';
 
  SWITCH:
-  my ($result_as, @key_cols) 
-    = ref $args->{-result_as} ? @{$args->{-result_as}}
-                              : ($args->{-result_as} || "rows");
+  my ($result_as, @subclass_args)
+    = does($arg_result_as, 'ARRAY') ? @$arg_result_as : ($arg_result_as);
+
+  # historically,some kinds of results accepted various aliases
+  $result_as =~ s/^flat(?:_array|)$/flat_arrayref/;
+  $result_as =~ s/^arrayref$/rows/;
+  $result_as =~ s/^fast-statement$/fast_statement/;
+
   for ($result_as) {
-
-    # CASE statement : the DBIx::DataModel::Statement object 
-    /^statement$/i and do {
-        delete $self->{args}{-result_as};
-        return $self;
-      };
-
-    # for all other cases, must first sqlize the statement
-    $self->sqlize if $self->{status} < SQLIZED;
-
-    # CASE sql : just return the SQL and bind values
-    /^sql$/i        and do {
-      not $callbacks 
-        or croak "$callbacks incompatible with -result_as=>'sql'";
-      return $self->sql;
-    };
-
-    # CASE subquery : return a ref to an arrayref with SQL and bind values
-    /^subquery$/i        and do {
-      not $callbacks 
-        or croak "$callbacks incompatible with -result_as=>'subquery'";
-      my ($sql, @bind) = $self->sql;
-      return \ ["($sql)", @bind];
-    };
-
-    # for all other cases, must first execute the statement
-    $self->execute;
-
-    # CASE sth : return the DBI statement handle
-    /^sth$/i        and do {
-        not $args->{-post_bless}
-          or croak "-post_bless incompatible with -result_as=>'sth'";
-        return $self->{sth};
-      };
-
-    # CASE rows : all data rows (this is the default)
-    /^(rows|arrayref)$/i  and return $self->all;
-
-    # CASE firstrow : just the first row
-    /^firstrow$/i   and return $self->next;
-
-    # CASE hashref : all data rows, put into a hashref
-    /^hashref$/i   and do {
-      @key_cols or @key_cols = $self->meta_source->primary_key
-        or croak "-result_as=>'hashref' impossible: no primary key";
-      my %hash;
-      while (my $row = $self->next) {
-        my @key;
-        foreach my $col (@key_cols) {
-          my $val = $row->{$col};
-          $val = '' if not defined $val; # $val might be 0, so no '||'
-          push @key, $val;
-        }
-        my $last_key_item = pop @key;
-        my $node          = \%hash;
-        $node = $node->{$_} ||= {} foreach @key;
-        $node->{$last_key_item} = $row;
-      }
-      return \%hash;
-    };
-
-    # CASE fast_statement : creates a reusable row
-    /^fast[-_]statement$/i and do {
-        $self->_build_reuse_row;
-        return $self;
-      };
-
-    # CASE flat_arrayref : flattened columns from each row
-    /^flat(?:_array(?:ref)?)?$/ and do {
-      $self->_build_reuse_row;
-      my @vals;
-      my $hash_key_name = $self->{sth}{FetchHashKeyName} || 'NAME';
-      my $cols = $self->{sth}{$hash_key_name};
-      while (my $row = $self->next) {
-        push @vals, @{$row}{@$cols};
-      }
-      return \@vals;
-    };
-
-
-    # OTHERWISE
-    croak "unknown -result_as value: $_"; 
+    my $subclass = $cache_result_class{$_}
+               //= $self->_find_result_class($_)
+      or croak "didn't find any ResultAs subclass to implement -result_as => '$_'";
+    my $result_maker = $subclass->new(@subclass_args);
+    return $result_maker->get_result($self);
   }
 }
 
@@ -506,21 +471,49 @@ sub row_count {
   my ($self) = @_;
 
   if (! exists $self->{row_count}) {
-    $self->sqlize if $self->{status} < SQLIZED;
+    $self->sqlize if $self->status < SQLIZED;
     my ($sql, @bind) = $self->sql;
-    $sql =~ s[^SELECT\b.*?\bFROM\b][SELECT COUNT(*) FROM]i
-      or croak "can't count rows from sql: $sql";
 
-    # TODO : line below is NOT PORTABLE. Should accomodate for all
-    # limit-offset syntaxes !!!
-    $sql =~ s[\bLIMIT \? OFFSET \?][]i
-      and splice @bind, -2;
+    # get syntax used for LIMIT clauses ...
+    my $sqla = $self->schema->sql_abstract;
+    my ($limit_sql, undef, undef) = $sqla->limit_offset(0, 0);
+    $limit_sql =~ s/([()?*])/\\$1/g;
 
+    # ...and use it to remove the LIMIT clause and associated bind vals, if any
+    if ($limit_sql =~ /ROWNUM/) { # special case for Oracle syntax, complex ...
+                                  # see source code of SQL::Abstract::More
+      $limit_sql =~ s/%s/(.*)/;
+      if ($sql =~ s/^$limit_sql/$1/) {
+        splice @bind, -2;
+      }
+    }
+    elsif ($sql =~ s[\b$limit_sql][]i) { # regular LIMIT/OFFSET syntaxes
+      splice @bind, -2;
+    }
+
+    # decide if the SELECT COUNT should wrap the original SQL in a subquery;
+    # this is needed with clauses like below that change the number of rows
+    my $should_wrap = $sql =~ /\b(UNION|INTERSECT|MINUS|EXCEPT|DISTINCT)\b/i;
+
+    # if no wrap required, attempt to directly substitute COUNT(*) for the 
+    # column names ...but if it fails, wrap anyway
+    $should_wrap ||= ! ($sql =~ s[^SELECT\b.*?\bFROM\b][SELECT COUNT(*) FROM]i);
+
+    # wrap SQL if needed, using  a subquery alias because it's required for 
+    # some DBMS (like PostgreSQL)
+    $should_wrap and  $sql = "SELECT COUNT(*) FROM "
+                           . $sqla->table_alias("( $sql )", "count_wrapper");
+
+    # log the statement and bind values
+    $self->schema->_debug("PREPARE $sql / @bind");
+
+    # call the database
     my $dbh    = $self->schema->dbh or croak "Schema has no dbh";
     my $method = $self->schema->dbi_prepare_method;
     my $sth    = $dbh->$method($sql);
     $sth->execute(@bind);
     ($self->{row_count}) = $sth->fetchrow_array;
+    $sth->finish;
   }
 
   return $self->{row_count};
@@ -535,9 +528,9 @@ sub row_num {
 sub next {
   my ($self, $n_rows) = @_;
 
-  $self->execute if $self->{status} < EXECUTED;
+  $self->execute if $self->status < EXECUTED;
 
-  my $sth      = $self->{sth}          or croak "absent sth in statement";
+  my $sth      = $self->sth            or croak "absent sth in statement";
   my $callback = $self->{row_callback} or croak "absent callback in statement";
 
   if (not defined $n_rows) {  # if user wants a single row
@@ -562,26 +555,20 @@ sub next {
     $self->{row_num} += @rows;
     return \@rows;
   }
+
+  # NOTE: ->next() returns a $row, while ->next(1) returns an arrayref of 1 row
 }
 
+
 sub all {
   my ($self) = @_;
 
-  $self->execute if $self->{status} < EXECUTED;
-
-  my $sth      = $self->{sth}          or croak "absent sth in statement";
-  my $callback = $self->{row_callback} or croak "absent callback in statement";
-
-  not $self->{reuse_row}  or croak "reusable row, cannot retrieve several";
-  my $rows = $sth->fetchall_arrayref({});
-  $callback->($_) foreach @$rows;
-  $self->{row_num} += @$rows;
-
-  return $rows;
+  # just call next() with a huge number
+  return $self->_next_and_finish(POSIX::LONG_MAX);
 }
 
 
-sub page_size   { shift->{args}{-page_size}  || POSIX::INT_MAX   }
+sub page_size   { shift->{args}{-page_size}  || POSIX::LONG_MAX  }
 sub page_index  { shift->{args}{-page_index} || 1                }
 sub offset      { shift->{offset}            || 0                }
 
@@ -595,37 +582,6 @@ sub page_count {
   return int(($row_count - 1) / $page_size) + 1;
 }
 
-sub goto_page {
-  my ($self, $page_index) = @_;
-
-  # if negative index, count down from last page
-  $page_index += $self->page_count + 1    if $page_index < 0;
-
-  $page_index >= 1 or croak "illegal page_index: $page_index";
-
-  $self->{page_index} = $page_index;
-  $self->{offset}     = ($page_index - 1) * $self->page_size;
-  $self->execute     unless $self->{row_num} == $self->{offset};
-
-  return $self;
-}
-
-
-sub shift_pages {
-  my ($self, $delta) = @_;
-
-  my $page_index = $self->page_index + $delta;
-  $page_index >= 1 or croak "illegal page index: $page_index";
-
-  $self->goto_page($page_index);
-}
-
-sub next_page {
-  my ($self) = @_;
-
-  $self->shift_pages(1);
-}
-
 
 sub page_boundaries {
   my ($self) = @_;
@@ -638,15 +594,18 @@ sub page_boundaries {
 
 sub page_rows {
   my ($self) = @_;
-  return $self->next($self->page_size);
+  return $self->_next_and_finish($self->page_size);
 }
 
 
 sub bless_from_DB {
   my ($self, $row) = @_;
 
-  # inject ref to $schema if in multi-schema mode
-  $row->{__schema} = $self->schema unless $self->schema->{is_singleton};
+  # inject ref to $schema if in multi-schema mode or if temporary
+  # db_schema is set
+  my $schema = $self->schema;
+  $row->{__schema} = $schema unless $schema->{is_singleton}
+                                 && !$schema->{db_schema};
 
   # bless into appropriate class
   bless $row, $self->meta_source->class;
@@ -662,25 +621,60 @@ sub bless_from_DB {
 }
 
 
+sub headers {
+  my $self = shift;
+
+  $self->status == EXECUTED
+    or $self->execute(@_);
+
+  my $hash_key_name = $self->sth->{FetchHashKeyName} || 'NAME';
+  return @{$self->sth->{$hash_key_name}};
+}
+
+
+sub finish {
+  my $self = shift;
+  $self->sth->finish;
+}
+
+
+sub make_fast {
+  my ($self) = @_;
+
+  $self->status == EXECUTED
+    or croak "cannot make_fast() when in state " . $self->status;
+
+  # create a reusable hash and bind_columns to it (see L<DBI/bind_columns>)
+  my %row;
+  $self->sth->bind_columns(\(@row{$self->headers}));
+  $self->{reuse_row} = \%row;
+}
+
+
 #----------------------------------------------------------------------
 # PRIVATE METHODS IN RELATION WITH SELECT()
 #----------------------------------------------------------------------
 
-sub _build_reuse_row {
-  my ($self) = @_;
+sub _forbid_callbacks {
+  my ($self, $subclass) = @_;
 
-  $self->{status} == EXECUTED
-    or croak "cannot _build_reuse_row() when in state $self->{status}";
-
-  # create a reusable hash and bind_columns to it (see L<DBI/bind_columns>)
-  my %row;
-  my $hash_key_name = $self->{sth}{FetchHashKeyName} || 'NAME';
-  $self->{sth}->bind_columns(\(@row{@{$self->{sth}{$hash_key_name}}}));
-  $self->{reuse_row} = \%row; 
+  my $callbacks = CORE::join ", ", grep {$self->arg($_)} 
+                                        qw/-pre_exec -post_exec -post_bless/;
+  if ($callbacks) {
+    $subclass =~ s/^.*:://;
+    croak "$callbacks incompatible with -result_as=>'$subclass'";
+  }
 }
 
 
 
+sub _next_and_finish {
+  my $self = shift;
+  my $row_or_rows = $self->next( @_ ); # pass original parameters
+  $self->finish;
+  return $row_or_rows;
+}
+
 sub _compute_from_DB_handlers {
   my ($self) = @_;
   my $meta_source    = $self->meta_source;
@@ -691,7 +685,11 @@ sub _compute_from_DB_handlers {
   # iterate over aliased_columns
   while (my ($alias, $column) = each %{$self->{aliased_columns} || {}}) {
     my $table_name;
-    $column =~ s/^(.+)\.// and $table_name = $1;
+    $column =~ s{^([^()]+)     # supposed table name (without parens)
+                  \.           # followed by a dot
+                  (?=[^()]+$)  # followed by supposed col name (without parens)
+                }{}x
+      and $table_name = $1;
     if (!$table_name) {
       $handlers{$alias} = $handlers{$column};
     }
@@ -699,10 +697,10 @@ sub _compute_from_DB_handlers {
       $table_name = $aliased_tables{$table_name} || $table_name;
 
       my $table   = $meta_schema->table($table_name)
-                 || firstval {($_->{db_name} || '') eq $table_name}
-                             ($meta_source, $meta_source->ancestors)
-                 # THINK: might perform a case-insensitive search 
-                 # (as second pass)
+                 || (firstval {($_->{db_name} || '') eq $table_name}
+                              ($meta_source, $meta_source->ancestors))
+                 || (firstval {uc($_->{db_name} || '') eq uc($table_name)}
+                              ($meta_source, $meta_source->ancestors))
         or croak "unknown table name: $table_name";
 
       $handlers{$alias} = $table->{column_handlers}->{$column};
@@ -712,7 +710,7 @@ sub _compute_from_DB_handlers {
   # handlers may be overridden from args{-column_types}
   if (my $col_types = $self->{args}{-column_types}) {
     while (my ($type_name, $columns) = each %$col_types) {
-      ref $columns or $columns = [$columns];
+      $columns = [$columns] unless does $columns, 'ARRAY';
       my $type = $self->schema->metadm->type($type_name)
         or croak "no such column type: $type_name";
       $handlers{$_} = $type->{handlers} foreach @$columns;
@@ -731,6 +729,30 @@ sub _compute_from_DB_handlers {
 }
 
 
+sub _find_result_class {
+  my $self         = shift;
+  my $name         = ucfirst shift;
+  my $schema       = $self->schema;
+  my $schema_class = ref $schema || $schema;
+
+  # try to find subclass $name within namespace of schema or ancestors
+  foreach my $namespace (@{mro::get_linear_isa($schema_class)}) {
+    my $class = "${namespace}::ResultAs::${name}";
+
+    # see if that class is already loaded (by checking for a 'get_result' method)
+    my $is_loaded  = defined &{$class."::get_result"};
+
+    # otherwise, try to load the module
+    $is_loaded ||= try   {load $class; 1}
+                   catch {die $_ if $_ !~ /^Can't locate(?! object method)/};
+
+    return $class if $is_loaded; # true : class is found, exit loop
+  }
+
+  return; # false : class not found
+}
+
+
 
 1; # End of DBIx::DataModel::Statement
 
@@ -754,20 +776,8 @@ the manual (purpose, lifecycle, etc.).
 
 =head1 METHODS
 
-=head2 new
-
-  my $statement 
-    = DBIx::DataModel::Statement->new($connected_source, %options);
-
-This is the statement constructor; C<$connected_source> is an
-instance of L<DBIx::DataModel::ConnectedSource>. 
-If present, C<%options> are delegated
-to the L<refine()|DBIx::DataModel::Doc::Reference/refine()> method.
-
-Explicit calls to the statement constructor are exceptional;
-the usual way to create a statement is through 
-L<ConnectedSource::select()|DBIx::DataModel::Doc::Reference/ConnectedSource::select()>.
-
+Methods for statements are described in the 
+L<Reference manual|DBIx::DataModel::Doc::Reference/"STATEMENTS">.
 
 =head1 PRIVATE METHOD NAMES
 
